#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 rdr_mqtt_sub.py

 @author East Japan Accounting Center Co., Ltd

 (c)2021 The University of Aizu
 This software is released under the New BSD License.
"""
import json
from logging import getLogger
import threading
import time

import paho.mqtt.client as mqtt

logger = getLogger(__name__)

DEBUG_MODE = True
MQTT_SETTING_PATH = "./config/mqtt_settings.json"


class RdrMqtt(mqtt.Client):

    def __init__(self):
        super(RdrMqtt, self).__init__()
        self._sub_topic = ""
        self._sub_topics = []
        self._mqtt_broker_connecting = False
        self._is_receive = False
        self._receive_topic_name = ""
        self._receive_message = None
        self.CONNECT_RETRY_CNT = 3
        self._RETRY_PUBLISH_MAX = 3
        self._lock = threading.Lock()
        self._PUBLISH_WAIT = 5

    def on_connect(self, client, obj, flags, rc):
        """
        Called when the broker responds to our connection request.
        See detail https://github.com/eclipse/paho.mqtt.python#callbacks

        Parameters
        ----------
        client
            the client instance for this callback
        obj
            the private user data as set in Client() or user_data_set()
        flags
            response flags sent by the broker
        rc
            the connection result
            0: Connection successful
            1: Connection refused - incorrect protocol version
            2: Connection refused - invalid client identifier
            3: Connection refused - server unavailable
            4: Connection refused - bad username or password
            5: Connection refused - not authorised
            6-255: Currently unused.

        Returns
        -------
        None
        """
        logger.info("Connected with result code %s", str(rc))
        if rc == 0:
            self._mqtt_broker_connecting = True
        else:
            self._mqtt_broker_connecting = False

    def on_message(self, client, obj, msg):
        """
        Called when a message has been received on a topic that the client subscribes to.
        See detail https://github.com/eclipse/paho.mqtt.python#callbacks

        Parameters
        ----------
        client
            the client instance for this callback
        obj
            the private user data as set in Client() or user_data_set()
        msg
            an instance of MQTTMessage

        Returns
        -------
        None
        """
        # logger.info(msg.topic + " " + str(msg.qos) + " " + str(msg.payload))
        # dict_message = json.loads(msg.payload)
        dict_message = json.loads(msg.payload.decode("utf-8"))
        self._lock.acquire()
        self._is_receive = True
        self._receive_topic_name = msg.topic
        self._receive_message = dict_message
        logger.info("on_message: {}".format(dict_message))
        self._lock.release()

    def on_subscribe(self, client, obj, mid, granted_qos):
        """
        Called when the broker responds to a subscribe request.
        See detail https://github.com/eclipse/paho.mqtt.python#callbacks

        Parameters
        ----------
        client
            the client instance for this callback
        obj
            the private user data as set in Client() or user_data_set()
        mid
            See above URL
        granted_qos
            See above URL

        Returns
        -------
        None
        """
        logger.info("Subscribed: " + str(mid) + " " + str(granted_qos))

    def on_disconnect(self, client, obj, rc):
        """
        Called when the client disconnects from the broker.
        See detail https://github.com/eclipse/paho.mqtt.python#callbacks

        Parameters
        ----------
        client
            the client instance for this callback
        obj
            the private user data as set in Client() or user_data_set()
        rc
            the connection result
            See detail on_connect()

        Returns
        -------
        None
        """
        logger.error("Disconnected. Result code %s", str(rc))
        self.loop_stop()
        self._mqtt_broker_connecting = False

    def publish_send(self, message, topic=""):
        """
        MQTTメッセージ送信を行う

        Parameters
        ----------
        message: Publish メッセージ
        topic:  Publishするトピック名

        Returns
        ----------
        result: 実行結果
        """
        result = False
        if not topic:
            return result
        # Convert message to Json format
        json_message = json.dumps(message)
        logger.debug("send topic: {}  message: {}".format(topic, json_message))
        for retry_count in range(self._RETRY_PUBLISH_MAX):
            # Publish message
            res = self.publish(topic, json_message)
            # Result judgment
            if res[0] == mqtt.MQTT_ERR_SUCCESS:
                result = True
                break
            else:
                time.sleep(self._PUBLISH_WAIT)
        return result

    def _mqtt_broker_connect(self):
        """
        Connect to specified MQTT Broker.

        Returns
        -------
        Connect result
        True: Success False: Failure
        """

        result = False
        retry_count = 0

        parse_result, param = self.is_json_file_parse(MQTT_SETTING_PATH)
        if parse_result is False:
            return result

        try:
            if DEBUG_MODE:
                pass
            else:
                self.tls_set(ca_certs=param['ca_certs'], certfile=param['certfile'], keyfile=param['keyfile'], cert_reqs=mqtt.ssl.CERT_REQUIRED, tls_version=mqtt.ssl.PROTOCOL_TLSv1_2, ciphers=None)
            while not self._mqtt_broker_connecting and retry_count < self.CONNECT_RETRY_CNT:
                self.connect(param['host'], param['port'])
                self.loop_start()
                time.sleep(1)
                if self._mqtt_broker_connecting:
                    result = True
                    break
                else:
                    retry_count += 1
                    logger.info("MQTT Broker now connecting...")
        except Exception:
            logger.error("Unable to MQTT connection")

        return result

    @staticmethod
    def is_json_file_parse(path):
        """
        Connect to specified MQTT Broker.

        Parameters
        ----------
        path
            json file path

        Returns
        -------
        Parse result
        True: Success False: Failure
        """

        result = False
        json_dict = {}

        try:
            with open(path) as f:
                json_dict = json.load(f)
                if "" not in json_dict.values():
                    result = True
                else:
                    logger.error("Param is empty.Please confirm.")
        except Exception:
            logger.error("Unable to read setting file.")

        return result, json_dict

    def _receive_param_init(self):
        """
        Receive Parameter Initialization

        Returns
        -------
        None
        """
        self._lock.acquire()
        self._is_receive = False
        self._receive_topic_name = ""
        self._receive_message = None
        self._lock.release()

    def subscribe_register(self, topic=None, topics=None):
        """
        Connect to specified MQTT Broker.
        Wait until receive the specific topic

        Parameters
        ----------
        topic
            Specify the subscribe topic

        topics
            Specify the subscribe topics

        Returns
        -------
        None
        """
        result = False
        if topic is None and topics is None:
            return result
        if self._mqtt_broker_connect():
            if topic is not None:
                self._sub_topic = topic
                self.subscribe(topic, 0)
                result = True
            elif topics is not None:
                if 1 < len(topics):
                    self._sub_topics = topics
                    topic_list = [(set_topic, 0) for set_topic in topics]
                    self.subscribe(topic_list)
                else:
                    self._sub_topic = topics[0]
                    self.subscribe(topics[0], 0)
                result = True
            else:
                pass
        return result

    def public_connection(self):
        """
        Connect to specified MQTT Broker.
        Wait until receive the specific topic

        Parameters
        ----------
        topic
            Specify the subscribe topic

        topics
            Specify the subscribe topics

        Returns
        -------
        None
        """
        result = False
        if self._mqtt_broker_connect():
            result = True
        return result


if __name__ == "__main__":
    pass
    # mqttc = RdrMqttSub()
    # mqttc.run("obj_recognition_01/#")
